@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2017  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2017-2019  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -140,10 +140,25 @@ static void interface_update_properties(struct proxy_interface *proxy,
 		proxy_interface_property_update(proxy, name, NULL);
 }
 
+static char *strdup_quoted_for_spaces(const char *str)
+{
+	const char *p;
+
+	for (p = str; *p; p++) {
+		if (*p != ' ')
+			continue;
+
+		return l_strdup_printf("\"%s\"", str);
+	}
+
+	return l_strdup(str);
+}
+
 char *proxy_property_str_completion(const struct proxy_interface_type *type,
 					proxy_property_match_func_t function,
 					const char *property_name,
-					const void *value, int state)
+					const void *value, int state,
+					const char *extra_interface)
 {
 	static struct l_queue *match;
 	static const struct l_queue_entry *entry;
@@ -163,13 +178,22 @@ char *proxy_property_str_completion(const struct proxy_interface_type *type,
 
 		entry = entry->next;
 
+		if (extra_interface) {
+			const char *path = proxy_interface_get_path(proxy);
+			const struct proxy_interface *extra =
+				proxy_interface_find(extra_interface, path);
+
+			if (!extra)
+				continue;
+		}
+
 		str = proxy_interface_property_tostr(proxy, property_name);
 		if (!str)
-			return NULL;
+			goto done;
 
-		return l_strdup(str);
+		return strdup_quoted_for_spaces(str);
 	}
-
+done:
 	l_queue_destroy(match, NULL);
 	match = NULL;
 	entry = NULL;
@@ -177,9 +201,9 @@ char *proxy_property_str_completion(const struct proxy_interface_type *type,
 	return NULL;
 }
 
-static char *proxy_property_completion_value_options(
-				const struct property_value_options *options,
-				const char *text, int state)
+static char *proxy_property_completion_value_options(const char **options,
+							const char *text,
+							int state)
 {
 	static int index;
 	static int len;
@@ -190,7 +214,7 @@ static char *proxy_property_completion_value_options(
 		len = strlen(text);
 	}
 
-	while ((opt = options[index++].value_str)) {
+	while ((opt = options[index++])) {
 		if (strncmp(opt, text, len))
 			continue;
 
@@ -401,7 +425,7 @@ struct l_queue *proxy_interface_find_all(const char *interface,
 	const struct l_queue_entry *entry;
 	struct l_queue *match = NULL;
 
-	if (!interface || !function)
+	if (!interface)
 		return NULL;
 
 	for (entry = l_queue_get_entries(proxy_interfaces); entry;
@@ -411,7 +435,7 @@ struct l_queue *proxy_interface_find_all(const char *interface,
 		if (!interface_match_by_type_name(proxy->type, interface))
 			continue;
 
-		if (!function(proxy->data, value))
+		if (function && !function(proxy->data, value))
 			continue;
 
 		if (!match)
@@ -423,26 +447,6 @@ struct l_queue *proxy_interface_find_all(const char *interface,
 	return match;
 }
 
-static struct l_queue *proxy_interface_find_by_path(const char *path)
-{
-	const struct l_queue_entry *entry;
-	struct l_queue *match = NULL;
-
-	for (entry = l_queue_get_entries(proxy_interfaces); entry;
-							entry = entry->next) {
-		struct proxy_interface *proxy = entry->data;
-
-		if (!strcmp(proxy->path, path)) {
-			if (!match)
-				match = l_queue_new();
-
-			l_queue_push_tail(match, proxy);
-		}
-	}
-
-	return match;
-}
-
 bool proxy_interface_is_same(const struct proxy_interface *a,
 					const struct proxy_interface *b)
 {
@@ -479,83 +483,6 @@ static void properties_changed_callback(struct l_dbus_message *message,
 	interface_update_properties(proxy, &changed, &invalidated);
 }
 
-static void proxy_interface_bind_dependencies(const char *path)
-{
-	const struct l_queue_entry *entry;
-	const struct l_queue_entry *inner_entry;
-	struct l_queue *match = proxy_interface_find_by_path(path);
-
-	if (l_queue_length(match) < 2)
-		goto done;
-
-	for (entry = l_queue_get_entries(match); entry; entry = entry->next) {
-		struct proxy_interface *proxy = entry->data;
-
-		if (!proxy->type->ops || !proxy->type->ops->bind_interface)
-			continue;
-
-		for (inner_entry = l_queue_get_entries(match); inner_entry;
-					inner_entry = inner_entry->next) {
-			char *error;
-			struct proxy_interface *dependency = inner_entry->data;
-
-			if (!strcmp(proxy->type->interface,
-						dependency->type->interface))
-				continue;
-
-			if (proxy->type->ops->bind_interface(proxy,
-								dependency))
-				continue;
-
-			error = l_strdup_printf("Interface %s does not support "
-						"dependency %s\n",
-						proxy->type->interface,
-						dependency->type->interface);
-			display_error(error);
-			l_free(error);
-		}
-	}
-
-done:
-	l_queue_destroy(match, NULL);
-}
-
-static void proxy_interface_unbind_dependencies(
-					const struct proxy_interface *proxy)
-{
-	const struct l_queue_entry *entry;
-	struct l_queue *match = proxy_interface_find_by_path(proxy->path);
-
-	if (l_queue_length(match) < 2)
-		goto done;
-
-	for (entry = l_queue_get_entries(match); entry; entry = entry->next) {
-		struct proxy_interface *dependency = entry->data;
-		char *error;
-
-		if (!strcmp(proxy->type->interface,
-						dependency->type->interface))
-			continue;
-
-		if (!dependency->type->ops ||
-				!dependency->type->ops->unbind_interface)
-			continue;
-
-		if (dependency->type->ops->unbind_interface(dependency, proxy))
-			continue;
-
-		error = l_strdup_printf("Interface %s does not support "
-					"dependency %s\n",
-					dependency->type->interface,
-					proxy->type->interface);
-		display_error(error);
-		l_free(error);
-	}
-
-done:
-	l_queue_destroy(match, NULL);
-}
-
 static bool is_ignorable(const char *interface)
 {
 	size_t i;
@@ -621,8 +548,6 @@ static void proxy_interface_create(const char *path,
 
 		l_queue_push_tail(proxy_interfaces, proxy);
 	}
-
-	proxy_interface_bind_dependencies(path);
 }
 
 static void proxy_interface_destroy(void *data)
@@ -677,6 +602,11 @@ const char *proxy_interface_get_interface(const struct proxy_interface *proxy)
 	return proxy->type->interface;
 }
 
+const char *proxy_interface_get_path(const struct proxy_interface *proxy)
+{
+	return proxy->path;
+}
+
 const char *proxy_interface_get_identity_str(
 					const struct proxy_interface *proxy)
 {
@@ -737,8 +667,6 @@ static void interfaces_removed_callback(struct l_dbus_message *message,
 		if (!proxy)
 			continue;
 
-		proxy_interface_unbind_dependencies(proxy);
-
 		l_queue_remove(proxy_interfaces, proxy);
 
 		proxy_interface_destroy(proxy);
@@ -820,6 +748,8 @@ static void service_disappeared_callback(struct l_dbus *dbus,
 
 	l_queue_clear(proxy_interfaces, proxy_interface_destroy);
 
+	command_reset_default_entities();
+
 	display_disable_cmd_prompt();
 }
 
@@ -828,8 +758,6 @@ static void dbus_disconnect_callback(void *user_data)
 	if (!command_is_interactive_mode())
 		return;
 
-	display("D-Bus disconnected, quitting...\n");
-
 	l_main_quit();
 }
 
